import logging

from .base import BaseRule

logger = logging.getLogger(__name__)


class PercentageFailuresRule(BaseRule):
    def __init__(
        self,
        max_failures_percentage: int,
        failure_cache_key: str,
        min_accepted_requests: int,
        request_cache_key: str,
    ) -> None:
        super().__init__(
            failure_cache_key=failure_cache_key,
            request_cache_key=request_cache_key,
        )
        self.max_failures_percentage = max_failures_percentage
        self.min_accepted_requests = min_accepted_requests

    def _get_percentage_failures(
        self, total_failures: int, total_requests: int
    ) -> float:
        if total_requests > 0:
            return (total_failures * 100) / total_requests
        return 0

    def should_open_circuit(
        self, total_failures: int, total_requests: int
    ) -> bool:
        percentage_failures = self._get_percentage_failures(
            total_failures=total_failures, total_requests=total_requests
        )
        return (
            total_requests > self.min_accepted_requests
            and percentage_failures >= self.max_failures_percentage
        )

    def log_increase_failures(
        self, total_failures: int, total_requests: int
    ) -> None:
        percentage_failures = self._get_percentage_failures(
            total_failures=total_failures, total_requests=total_requests
        )
        logger.info(
            f'Increase failure for: {self.failure_cache_key} - '
            f'max failures {self.max_failures_percentage}% - '
            f'total failures {total_failures} - '
            f'min accepted requests {self.min_accepted_requests} - '
            f'total requests {total_requests} - '
            f'percentage failures {percentage_failures}%'
        )



from lasier.types import Timeout

from .base import CacheAdapterBase


class Adapter(CacheAdapterBase):
    def add(self, key: str, value: int, timeout: Timeout = None) -> None:
        self.cache.set(key, value, timeout, nx=True)



from lasier.types import Timeout

from .base import CacheAdapterBase


class Adapter(CacheAdapterBase):
    def expire(self, key: str, timeout: Timeout = None) -> None:
        self.cache.touch(key, timeout)



import inspect
import logging
from typing import Any, Iterable, Optional, Type

from lasier.types import Timeout

from .rules.base import BaseRule

logger = logging.getLogger(__name__)
_ONE_MINUTE = 60


class CircuitBreakerBase:
    def __init__(
        self,
        rule: BaseRule,
        cache: Any,
        failure_exception: Type[Exception],
        failure_timeout: Timeout = _ONE_MINUTE,
        circuit_timeout: Timeout = _ONE_MINUTE,
        catch_exceptions: Optional[Iterable[Type[Exception]]] = None,
    ) -> None:
        self.rule = rule
        self.cache = cache
        self.failure_timeout = failure_timeout
        self.circuit_timeout = circuit_timeout
        self.circuit_cache_key = 'circuit_{}'.format(rule.failure_cache_key)
        self.failure_exception = failure_exception
        self.catch_exceptions = catch_exceptions or (Exception,)

    def _is_catchable_exception(self, exception: Type[Exception]) -> bool:
        return inspect.isclass(exception) and any(
            issubclass(exception, exception_class)
            for exception_class in self.catch_exceptions
        )

    def _notify_open_circuit(self) -> None:
        logger.critical(
            f'Open circuit for {self.rule.failure_cache_key} '
            f'{self.circuit_cache_key}'
        )

    def _notify_max_failures_exceeded(self) -> None:
        logger.info(f'Max failures exceeded by: {self.rule.failure_cache_key}')



import asyncio
from functools import wraps

from lasier.types import Timeout

from .base import CircuitBreakerBase


class CircuitBreaker(CircuitBreakerBase):
    async def is_circuit_open(self) -> bool:
        return await self.cache.get(self.circuit_cache_key) or False

    async def get_total_failures(self) -> int:
        return await self.cache.get(self.rule.failure_cache_key) or 0

    async def get_total_requests(self) -> int:
        return await self.cache.get(self.rule.request_cache_key) or 0

    async def open_circuit(self) -> None:
        await self.cache.set(self.circuit_cache_key, 1, self.circuit_timeout)

        # Delete the cache key to mitigate multiple sequentials openings
        # when a key is created accidentally without timeout (from an incr
        # operation)
        await asyncio.gather(
            self.cache.delete(self.rule.failure_cache_key),
            self.cache.delete(self.rule.request_cache_key),
        )

        self._notify_open_circuit()

    async def __aenter__(self):
        if await self.is_circuit_open():
            raise self.failure_exception

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._increase_request_count()
        if self._is_catchable_exception(exc_type):
            await self._increase_failure_count()

            total_failures, total_requests = await asyncio.gather(
                self.get_total_failures(), self.get_total_requests()
            )

            if self.rule.should_open_circuit(
                total_failures=total_failures, total_requests=total_requests
            ):
                await self.open_circuit()
                self._notify_max_failures_exceeded()
                raise self.failure_exception

    def __call__(self, func):
        @wraps(func)
        async def inner(*args, **kwargs):
            async with self:
                return await func(*args, **kwargs)

        return inner

    async def _increase_failure_count(self) -> None:
        if (
            await self.is_circuit_open()
            or not self.rule.should_increase_failure_count()
        ):
            return

        total_failures, total_requests = await asyncio.gather(
            self._incr(self.rule.failure_cache_key, self.failure_timeout),
            self.get_total_requests(),
        )

        self.rule.log_increase_failures(
            total_failures=total_failures, total_requests=total_requests
        )

    async def _increase_request_count(self) -> None:
        if (
            await self.is_circuit_open()
            or not self.rule.should_increase_request_count()
        ):
            return

        await self._incr(
            self.rule.request_cache_key, self.failure_timeout  # type: ignore
        )
        # To calculate the exact percentage, the cache of requests and the
        # cache of failures must expire at the same time.
        if self.rule.should_increase_failure_count():
            await self.cache.add(
                self.rule.failure_cache_key, 0, self.failure_timeout
            )

    async def _incr(self, key: str, timeout: Timeout) -> int:
        value = await self.cache.incr(key)
        if value == 1:
            await self.cache.expire(key, timeout)
        return value


circuit_breaker = CircuitBreaker



from lasier.types import Timeout

from .base import AsyncCacheAdapterBase


class Adapter(AsyncCacheAdapterBase):
    async def add(self, key: str, value: int, timeout: Timeout = None) -> None:
        try:
            await super().add(key, value, timeout)
        except ValueError:
            return

    async def incr(self, key: str) -> int:
        return await self.cache.increment(key)

    async def flushdb(self) -> None:
        await self.cache.clear()



import abc
from typing import Optional


class BaseRule(metaclass=abc.ABCMeta):
    def __init__(
        self, failure_cache_key: str, request_cache_key: Optional[str] = None
    ) -> None:
        self.failure_cache_key = failure_cache_key
        self.request_cache_key = request_cache_key

    @abc.abstractmethod
    def should_open_circuit(
        self, total_failures: int, total_requests: int
    ) -> bool:
        pass

    def should_increase_failure_count(self) -> bool:
        return self.failure_cache_key is not None

    def should_increase_request_count(self) -> bool:
        return self.request_cache_key is not None

    @abc.abstractmethod
    def log_increase_failures(
        self, total_failures: int, total_requests: int
    ) -> None:
        pass



import logging

from .base import BaseRule

logger = logging.getLogger(__name__)


class MaxFailuresRule(BaseRule):
    def __init__(self, max_failures: int, failure_cache_key: str) -> None:
        super().__init__(failure_cache_key=failure_cache_key)
        self.max_failures = max_failures

    def should_open_circuit(
        self, total_failures: int, total_requests: int
    ) -> bool:
        return total_failures >= self.max_failures

    def log_increase_failures(
        self, total_failures: int, total_requests: int
    ) -> None:
        logger.info(
            f'Increase failure for: {self.failure_cache_key} - '
            f'max failures {self.max_failures} - '
            f'total requests {total_requests} - '
            f'total failures {total_failures}'
        )



from typing import Any, Optional

from lasier.types import Timeout


class _CacheAdapterMixin:
    def __init__(self, cache: Any) -> None:
        self.cache = cache

    def _convert_to_int(self, value: Optional[str]) -> Optional[int]:
        if value is not None:
            return int(value)
        return None


class CacheAdapterBase(_CacheAdapterMixin):
    def add(self, key: str, value: int, timeout: Timeout = None) -> None:
        self.cache.add(key, value, timeout)

    def set(self, key: str, value: int, timeout: Timeout = None) -> None:
        self.cache.set(key, value, timeout)

    def incr(self, key: str) -> int:
        return self.cache.incr(key)

    def get(self, key: str) -> Optional[int]:
        return self._convert_to_int(self.cache.get(key))

    def expire(self, key: str, timeout: Timeout) -> None:
        self.cache.expire(key, timeout)

    def delete(self, key: str) -> None:
        self.cache.delete(key)

    def flushdb(self) -> None:
        self.cache.flushdb()


class AsyncCacheAdapterBase(_CacheAdapterMixin):
    async def add(self, key: str, value: int, timeout: Timeout = None) -> None:
        await self.cache.add(key, value, timeout)

    async def set(self, key: str, value: int, timeout: Timeout = None) -> None:
        await self.cache.set(key, value, timeout)

    async def incr(self, key: str) -> int:
        return await self.cache.incr(key)

    async def get(self, key: str) -> Optional[int]:
        return self._convert_to_int(await self.cache.get(key))

    async def expire(self, key: str, timeout: Timeout) -> None:
        await self.cache.expire(key, timeout)

    async def delete(self, key: str) -> None:
        await self.cache.delete(key)

    async def flushdb(self) -> None:
        await self.cache.flushdb()
